import os
import torch

def print_details():
    for slurm_var in ["SLURMD_NODENAME","SBATCH_CLUSTERS", 
                      "SBATCH_PARTITION","SLURM_JOB_PARTITION",
                      "SLURM_NODEID","SLURM_MEM_PER_GPU",
                      "SLURM_MEM_PER_CPU","SLURM_MEM_PER_NODE","SLURM_JOB_ID"]:
        try:
            print(slurm_var, os.environ[slurm_var])
        except:
            print(slurm_var, "doesnt exist")
    try:
        print('torch.cuda.get_device_name()',torch.cuda.get_device_name())
        print('torch.cuda.get_device_capability()',torch.cuda.get_device_capability())
        current_device = torch.cuda.current_device()
        gpu = torch.cuda.get_device_properties(current_device)
        print(f"GPU Name: {gpu.name}")
        print(f"GPU Memory Total: {gpu.total_memory / 1024**2} MB")
        print(f"GPU Memory Free: {torch.cuda.memory_allocated(current_device) / 1024**2} MB")
        print(f"GPU Memory Used: {torch.cuda.memory_reserved(current_device) / 1024**2} MB")
    except:
        print("couldnt print cuda details")


'''import torch
from torchvision import models, transforms
from PIL import Image
import requests
from io import BytesIO
import numpy as np

# Load the pre-trained DeepLabV3 model
model = models.segmentation.deeplabv3_resnet101(pretrained=True)
model.eval()

import torch
from torchvision import models, transforms
from PIL import Image
import requests
from io import BytesIO
import numpy as np

# Load the pre-trained DeepLabV3 model
model = models.segmentation.deeplabv3_resnet101(pretrained=True)
model.eval()

# Function to preprocess the image
def preprocess(image):
    preprocess_transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
    ])
    return preprocess_transform(image).unsqueeze(0)

# Function to perform segmentation
def segment_image(image):
    input_tensor = preprocess(image)
    with torch.no_grad():
        output = model(input_tensor)['out'][0]
    return output.argmax(0)

# Function to extract clothing pixels and create a new image with non-clothing pixels set to black
def create_clothing_only_image(image, segmentation):
    clothing_ids = [

    ]  # Adjust these IDs based on the dataset used
    mask = torch.zeros_like(segmentation, dtype=torch.uint8)
    for idx in clothing_ids:
        mask[segmentation == idx] = 1

    # Convert the mask to a numpy array and expand dimensions
    mask_np = mask.numpy().astype(np.uint8)
    mask_expanded = np.expand_dims(mask_np, axis=2)

    # Convert the original image to a numpy array
    image_np = np.array(image)

    # Apply the mask to the image
    clothing_only_image_np = image_np * mask_expanded

    # Convert the result back to a PIL image
    clothing_only_image = Image.fromarray(clothing_only_image_np.astype(np.uint8))
    return clothing_only_image
'''
import torch.nn.functional as F
from PIL import Image
from cloths_segmentation.pre_trained_models import create_model
from segmentation_models_pytorch import Unet
from segmentation_models_pytorch.decoders.unet.decoder import UnetDecoder

import numpy as np
from torchvision.transforms import PILToTensor,Normalize,ToPILImage
import torch

class BetterUnetDecoder(UnetDecoder):
    def __init__(self,parent:UnetDecoder):
        for key, value in parent.__dict__.items():
            setattr(self, key, value)

class BetterUnet(Unet):
    def __init__(self, parent:Unet,device:str, dtype:torch.dtype):
        
        parent=parent.to(device)
        parent=parent.to(dtype)
        for key, value in parent.__dict__.items():
            setattr(self, key, value)
        self.device=device
        self.dtype=dtype

    @torch.no_grad()
    def predict_embeds(self,x)->list:
        self.check_input_shape(x)

        features = self.encoder(x)
        for f in features:
            print(f.size())

    @torch.no_grad()
    def decoder_embeds(self,x)->torch.Tensor:
        self.check_input_shape(x)

        features = self.encoder(x)

def get_segmentation_model(device:str,dtype:torch.dtype)->BetterUnet:
    parent = create_model("Unet_2020-10-30")
    return BetterUnet(parent,device,dtype)

def pad_to_multiple_of_32(image_tensor: torch.Tensor):
    _, _, h, w = image_tensor.size()
    pad_h = (32 - h % 32) % 32
    pad_w = (32 - w % 32) % 32
    padding = (0, pad_w, 0, pad_h)  # left, right, top, bottom
    padded_image = F.pad(image_tensor, padding, mode='constant', value=0)
    return padded_image, padding

# Function to unpad the image tensor
def unpad_image(image_tensor:torch.Tensor, padding:tuple):
    _, _, h, w = image_tensor.size()
    pad_h, pad_w = padding[3], padding[1]
    unpadded_image = image_tensor[:, :, :h - pad_h, :w - pad_w]
    return unpadded_image

def process_image(image: Image.Image,device:str,dtype:torch.dtype)->tuple[torch.Tensor,torch.Tensor,tuple]:
    tensor_img=PILToTensor()(image)
    tensor_img=torch.unsqueeze(tensor_img,0)
    tensor_img,padding=pad_to_multiple_of_32(tensor_img)
    base_image=tensor_img
    tensor_img=tensor_img.to(dtype)
    tensor_img=Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))(tensor_img)
    
    #tensor_img,padding=pad_to_multiple_of_32(tensor_img)
    if device is not None:
        tensor_img=tensor_img.to(device)
    return tensor_img,base_image,padding


def get_mask(tensor_img:torch.Tensor,segmentation_model:BetterUnet,threshold:int)->torch.Tensor:
    prediction = segmentation_model(tensor_img)
    #segmentation_model.predict_embeds(tensor_img)
    #print(prediction.size())
    #print(torch.max(prediction),torch.min(prediction))
    #print(torch.mean(prediction),torch.max(prediction),torch.min(prediction))
    mask=(prediction > threshold).to(torch.uint8)
    return mask

@torch.no_grad()
def clothes_segmentation(image: Image.Image,segmentation_model:BetterUnet,threshold:int)->Image.Image:
    tensor_img,base_image,padding=process_image(image,segmentation_model.device,segmentation_model.dtype)
    mask=get_mask(tensor_img,segmentation_model,threshold)
    masked_tensor=base_image.to("cpu")*mask.to("cpu")
    unpadded_masked_tensor=unpad_image(masked_tensor,padding)
    return ToPILImage()(unpadded_masked_tensor.squeeze(0))



# coding=utf-8
# Copyright 2021 Google AI, Ross Wightman, The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" PyTorch ViT model."""



import math
from typing import Dict, List, Optional, Set, Tuple, Union

import torch
import torch.utils.checkpoint
from torch import nn

from transformers.activations import ACT2FN
from transformers.utils import ModelOutput
from transformers.modeling_outputs import (BaseModelOutput, BaseModelOutputWithPooling, 
    ImageClassifierOutput, MaskedImageModelingOutput,
    BaseModelOutputWithPast, BaseModelOutputWithPastAndCrossAttentions)
from transformers.modeling_utils import PreTrainedModel
from transformers.pytorch_utils import find_pruneable_heads_and_indices, prune_linear_layer
from transformers.models.vit.configuration_vit import ViTConfig
from transformers.models.vit.modeling_vit import (ViTEmbeddings, ViTPatchEmbeddings,ViTSelfOutput,ViTIntermediate,ViTOutput, ViTPooler)
from dataclasses import dataclass


@dataclass
class BaseModelOutputWithPastAndPooling(ModelOutput):
    """
    Base class for model's outputs that may also contain a past key/values (to speed up sequential decoding).

    Args:
        last_hidden_state (`torch.FloatTensor` of shape `(batch_size, sequence_length, hidden_size)`):
            Sequence of hidden-states at the output of the last layer of the model.

            If `past_key_values` is used only the last hidden-state of the sequences of shape `(batch_size, 1,
            hidden_size)` is output.
        pooler_output (`torch.FloatTensor` of shape `(batch_size, hidden_size)`):
            Last layer hidden-state of the first token of the sequence (classification token) after further processing
            through the layers used for the auxiliary pretraining task. E.g. for BERT-family of models, this returns
            the classification token after processing through a linear layer and a tanh activation function. The linear
            layer weights are trained from the next sentence prediction (classification) objective during pretraining.
        past_key_values (`tuple(tuple(torch.FloatTensor))`, *optional*, returned when `use_cache=True` is passed or when `config.use_cache=True`):
            Tuple of `tuple(torch.FloatTensor)` of length `config.n_layers`, with each tuple having 2 tensors of shape
            `(batch_size, num_heads, sequence_length, embed_size_per_head)`) 

            Contains pre-computed hidden-states (key and values in the self-attention blocks)
        hidden_states (`tuple(torch.FloatTensor)`, *optional*, returned when `output_hidden_states=True` is passed or when `config.output_hidden_states=True`):
            Tuple of `torch.FloatTensor` (one for the output of the embeddings, if the model has an embedding layer, +
            one for the output of each layer) of shape `(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the model at the output of each layer plus the optional initial embedding outputs.
        attentions (`tuple(torch.FloatTensor)`, *optional*, returned when `output_attentions=True` is passed or when `config.output_attentions=True`):
            Tuple of `torch.FloatTensor` (one for each layer) of shape `(batch_size, num_heads, sequence_length,
            sequence_length)`.

            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
    """

    last_hidden_state: torch.FloatTensor = None
    past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    hidden_states: Optional[Tuple[torch.FloatTensor, ...]] = None
    attentions: Optional[Tuple[torch.FloatTensor, ...]] = None
    pooler_output: torch.FloatTensor = None

class BetterViTSelfAttention(nn.Module):
    def __init__(self, config: ViTConfig) -> None:
        super().__init__()
        if config.hidden_size % config.num_attention_heads != 0 and not hasattr(config, "embedding_size"):
            raise ValueError(
                f"The hidden size {config.hidden_size,} is not a multiple of the number of attention "
                f"heads {config.num_attention_heads}."
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size, bias=config.qkv_bias)
        self.key = nn.Linear(config.hidden_size, self.all_head_size, bias=config.qkv_bias)
        self.value = nn.Linear(config.hidden_size, self.all_head_size, bias=config.qkv_bias)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

    def transpose_for_scores(self, x: torch.Tensor) -> torch.Tensor:
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
        self, hidden_states, head_mask: Optional[torch.Tensor] = None, output_attentions: bool = False, output_past_key_values: bool=False,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor]]:
        mixed_query_layer = self.query(hidden_states)

        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))
        query_layer = self.transpose_for_scores(mixed_query_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

        attention_scores = attention_scores / math.sqrt(self.attention_head_size)

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(new_context_layer_shape)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)
        outputs = (*outputs, (key_layer,value_layer )) if output_past_key_values else outputs

        return outputs



class BetterViTAttention(nn.Module):
    def __init__(self, config: ViTConfig) -> None:
        super().__init__()
        self.attention = BetterViTSelfAttention(config)
        self.output = ViTSelfOutput(config)
        self.pruned_heads = set()

    def prune_heads(self, heads: Set[int]) -> None:
        if len(heads) == 0:
            return
        heads, index = find_pruneable_heads_and_indices(
            heads, self.attention.num_attention_heads, self.attention.attention_head_size, self.pruned_heads
        )

        # Prune linear layers
        self.attention.query = prune_linear_layer(self.attention.query, index)
        self.attention.key = prune_linear_layer(self.attention.key, index)
        self.attention.value = prune_linear_layer(self.attention.value, index)
        self.output.dense = prune_linear_layer(self.output.dense, index, dim=1)

        # Update hyper params and store pruned heads
        self.attention.num_attention_heads = self.attention.num_attention_heads - len(heads)
        self.attention.all_head_size = self.attention.attention_head_size * self.attention.num_attention_heads
        self.pruned_heads = self.pruned_heads.union(heads)

    def forward(
        self,
        hidden_states: torch.Tensor,
        head_mask: Optional[torch.Tensor] = None,
        output_attentions: bool = False,
        output_past_key_values:bool=False
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor]]:
        self_outputs = self.attention(hidden_states, head_mask, output_attentions,output_past_key_values)

        attention_output = self.output(self_outputs[0], hidden_states)

        outputs = (attention_output,) + self_outputs[1:]  # add attentions if we output them
        return outputs


class BetterViTLayer(nn.Module):
    """This corresponds to the Block class in the timm implementation."""

    def __init__(self, config: ViTConfig) -> None:
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1
        self.attention = BetterViTAttention(config)
        self.intermediate = ViTIntermediate(config)
        self.output = ViTOutput(config)
        self.layernorm_before = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.layernorm_after = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)

    def forward(
        self,
        hidden_states: torch.Tensor,
        head_mask: Optional[torch.Tensor] = None,
        output_attentions: bool = False,
         output_past_key_values:bool=False
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor]]:
        self_attention_outputs = self.attention(
            self.layernorm_before(hidden_states),  # in ViT, layernorm is applied before self-attention
            head_mask,
            output_attentions=output_attentions,
            output_past_key_values=output_past_key_values
        )
        attention_output = self_attention_outputs[0]
        outputs = self_attention_outputs[1:]  # add self attentions/past_keys

        # first residual connection
        hidden_states = attention_output + hidden_states

        # in ViT, layernorm is also applied after self-attention
        layer_output = self.layernorm_after(hidden_states)
        layer_output = self.intermediate(layer_output)

        # second residual connection is done here
        layer_output = self.output(layer_output, hidden_states)

        outputs = (layer_output,) + outputs
        return outputs


class BetterViTEncoder(nn.Module):
    def __init__(self, config: ViTConfig) -> None:
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList([BetterViTLayer(config) for _ in range(config.num_hidden_layers)])
        self.gradient_checkpointing = False

    def forward(
        self,
        hidden_states: torch.Tensor,
        head_mask: Optional[torch.Tensor] = None,
        output_attentions: bool = False,
        output_hidden_states: bool = False,
         output_past_key_values:bool=False,
        return_dict: bool = True,
    ) -> Union[tuple, BaseModelOutput]:
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None
        all_past_key_values= () if output_past_key_values else None

        for i, layer_module in enumerate(self.layer):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None

            if self.gradient_checkpointing and self.training:
                layer_outputs = self._gradient_checkpointing_func(
                    layer_module.__call__,
                    hidden_states,
                    layer_head_mask,
                    output_attentions,
                    output_past_key_values
                )
            else:
                layer_outputs = layer_module(hidden_states, layer_head_mask, output_attentions,output_past_key_values)

            hidden_states = layer_outputs[0]

            if output_attentions and output_past_key_values:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)
                all_past_key_values =all_past_key_values+(layer_outputs[2],)
            elif output_attentions:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)
            elif output_past_key_values:
                all_past_key_values =all_past_key_values+(layer_outputs[1],)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, all_hidden_states, all_self_attentions] if v is not None)
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
            cross_attentions=None,
            past_key_values=all_past_key_values
        ) 


class BetterViTPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = ViTConfig
    base_model_prefix = "vit"
    main_input_name = "pixel_values"
    supports_gradient_checkpointing = True
    _no_split_modules = ["ViTEmbeddings", "ViTLayer"]

    def _init_weights(self, module: Union[nn.Linear, nn.Conv2d, nn.LayerNorm]) -> None:
        """Initialize the weights"""
        if isinstance(module, (nn.Linear, nn.Conv2d)):
            # Upcast the input in `fp32` and cast it back to desired `dtype` to avoid
            # `trunc_normal_cpu` not implemented in `half` issues
            module.weight.data = nn.init.trunc_normal_(
                module.weight.data.to(torch.float32), mean=0.0, std=self.config.initializer_range
            ).to(module.weight.dtype)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        elif isinstance(module, ViTEmbeddings):
            module.position_embeddings.data = nn.init.trunc_normal_(
                module.position_embeddings.data.to(torch.float32),
                mean=0.0,
                std=self.config.initializer_range,
            ).to(module.position_embeddings.dtype)

            module.cls_token.data = nn.init.trunc_normal_(
                module.cls_token.data.to(torch.float32),
                mean=0.0,
                std=self.config.initializer_range,
            ).to(module.cls_token.dtype)


class BetterViTModel(BetterViTPreTrainedModel):
    def __init__(self, config: ViTConfig, add_pooling_layer: bool = True, use_mask_token: bool = False):
        super().__init__(config)
        self.config = config

        self.embeddings = ViTEmbeddings(config, use_mask_token=use_mask_token)
        self.encoder = BetterViTEncoder(config)

        self.layernorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.pooler = ViTPooler(config) if add_pooling_layer else None

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self) -> ViTPatchEmbeddings:
        return self.embeddings.patch_embeddings

    def _prune_heads(self, heads_to_prune: Dict[int, List[int]]) -> None:
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    def forward(
        self,
        pixel_values: Optional[torch.Tensor] = None,
        bool_masked_pos: Optional[torch.BoolTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_past_key_values: Optional[bool] = None,
        interpolate_pos_encoding: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, BaseModelOutputWithPastAndPooling]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if pixel_values is None:
            raise ValueError("You have to specify pixel_values")

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        # TODO: maybe have a cleaner way to cast the input (from `ImageProcessor` side?)
        expected_dtype = self.embeddings.patch_embeddings.projection.weight.dtype
        if pixel_values.dtype != expected_dtype:
            pixel_values = pixel_values.to(expected_dtype)

        embedding_output = self.embeddings(
            pixel_values, bool_masked_pos=bool_masked_pos, interpolate_pos_encoding=interpolate_pos_encoding
        )

        encoder_outputs = self.encoder(
            embedding_output,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            output_past_key_values=output_past_key_values
        )
        sequence_output = encoder_outputs[0]
        sequence_output = self.layernorm(sequence_output)
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            head_outputs = (sequence_output, pooled_output) if pooled_output is not None else (sequence_output,)
            return head_outputs + encoder_outputs[1:]

        return BaseModelOutputWithPastAndPooling(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            past_key_values=encoder_outputs.past_key_values
        )



import os
import sys
sys.path.append('/home/jlb638/Desktop/package')
from src.experiment_helpers.better_ddpo_pipeline import BetterDefaultDDPOStableDiffusionPipeline
from src.experiment_helpers.lora_loading import save_lora_weights, load_lora_weights,save_pipeline_hf,get_pipeline_from_hf, fix_lora_weights
from peft import LoraConfig,PeftModel
from diffusers import StableDiffusionPipeline
from src.experiment_helpers.utils import print_trainable_parameters

lora_config = LoraConfig(
                r=4,
                lora_alpha=4,
                init_lora_weights="gaussian",
                target_modules=["to_k", "to_q", "to_v", "to_out.0"],)

fix_lora_weights("jlbaker361/ddpo_512_wikiart-subjects_dcgan")

pipeline=StableDiffusionPipeline.from_pretrained("stabilityai/stable-diffusion-2-base")
unet=pipeline.unet

print_trainable_parameters(unet)
unet=PeftModel.from_pretrained(unet,"jlbaker361/ddpo_512_wikiart-subjects_dcgan")
print_trainable_parameters(unet)
pipeline("prompt",num_inference_steps=2)



from collections import deque
from datasets import load_dataset
import itertools
import numpy as np
from PIL import Image,ImageDraw,ImageFont
import wandb

def get_run_id(project_name,file_path):
    run_id=None
    try:
        with open(file_path, 'r') as file:
            for line in file:
                if line.find(f"https://wandb.ai/jlbaker361/{project_name}/runs/")!=-1:
                    run_id=line[line.rfind("/")+1:].strip()
                if line.find("CANCELLED AT")!=-1:
                    print("Cancelled", file_path)
    except:
        pass
    if run_id is None:
        print("couldn't find run id!",file_path)
    return run_id

def get_grouping_dict(project_name:str,file_path_format:str,key_value_dict:dict, grouping_keys:list)-> dict:
    assert all(key in key_value_dict for key in grouping_keys)
    grouping_keys.sort()
    relevant_values=[key_value_dict[key] for key in grouping_keys]
    combinations = ["_".join(t) for t in  list(itertools.product(*relevant_values))]
    grouping_dict={
        c:[] for c in combinations
    }

    print(grouping_dict)

    all_values=[]
    for k,v_list in key_value_dict.items():
        all_values.append([(k,v) for v in v_list])
    all_combinations=list(itertools.product(*all_values))
    all_combo_dicts=[
        {key:value for (key,value) in combination} for combination in all_combinations
    ]
    
    for combo_dict in all_combo_dicts:
        #print(combo_dict)
        file_path=file_path_format
        for key,value in combo_dict.items():
            file_path=file_path.replace("{"+key+"}",value)
        run_id=get_run_id(project_name,file_path)
        
        group_id="_".join([combo_dict[key] for key in grouping_keys])
        print('\t',file_path,run_id,group_id)
        grouping_dict[group_id].append(run_id)
    
    return grouping_dict

def make_tables(run_id_dict:dict,key_list:list,project:str):
    api=wandb.Api(timeout=60)
    for name,run_id_list in run_id_dict.items():
        metric_dict={
            key:[] for key in key_list
        }
        for run_id in run_id_list:
            try:
                run=api.run(f"jlbaker361/{project}/{run_id}")
                for key in key_list:
                    try:
                        history=run.history(keys=[key])
                        metric_dict[key].append(history[key][0])
                    except:
                        print(f"couldnt find key {key} for name {name} and {run_id}")
            except:
                pass
        #print(metric_dict)
        print(name.replace("_", " "), "&", " & ".join([f"{round(np.mean(metric_dict[key]),4)}" for key in metric_dict.keys()]),"\\\\")


def make_grid(project_name:str,
              methods_to_target_dict:dict,
              target_column:str, 
              image_column:str,
              size:tuple,
              suffix_list:list,
              gd:dict,
              prompt_id:str)->list:
    api=wandb.Api(timeout=60)
    grid=[]
    data=load_dataset(f"jlbaker361/{project_name}",split="train")
    for method,target_list in methods_to_target_dict.items():
        for target in target_list:
            new_row=[]
            for line in data:
                if line[target_column]==target:
                    new_row.append(line[image_column].resize(size))
                    break
            for suffix in suffix_list:
                key=f"{method}_{suffix}"
                run_id_list= gd[key]
                for normal_run in run_id_list:
                    try:
                        run=api.run(f"jlbaker361/{project_name}/{normal_run}")
                        for file in run.files():
                            if file.mimetype=="image/png":
                                query_string=f"media/images/{target}/{method}_{prompt_id}"
                                if file.name.find(query_string)!=-1:
                                    image_path=file.name
                                    print(query_string, image_path)
                                    file.download(exist_ok=True)
                                    with Image.open(image_path) as img:
                                        # Resize image
                                        img = img.resize(size)
                                        new_row.append(img)
                                        break
                    except:
                        pass
            grid.append(new_row)
            print(target, new_row)
    return grid

def create_image_grid_with_labels(images, row_labels, col_labels, image_size, label_font_size=20,margin=10, horiz_margin=50):
    """
    Create a grid of images with axis labels.

    :param images: 2D list of PIL Image objects.
    :param row_labels: List of labels corresponding to the rows of images.
    :param col_labels: List of labels corresponding to the columns of images.
    :param image_size: Tuple (width, height) representing the size of each image.
    :param label_font_size: Font size for the labels.
    :return: PIL Image object representing the grid.
    """
    # Find unique row and column labels
    unique_row_labels = row_labels
    unique_col_labels = col_labels

    # Create mapping from labels to grid positions
    row_map = {label: idx for idx, label in enumerate(unique_row_labels)}
    col_map = {label: idx for idx, label in enumerate(unique_col_labels)}

    # Calculate the grid size
    num_rows = len(unique_row_labels)
    num_cols = len(unique_col_labels)

    img_width, img_height = image_size

    # Calculate the total size including labels
    label_padding = label_font_size + margin
    grid_width = (num_cols) * (img_width+margin) + label_padding+horiz_margin
    grid_height = (num_rows )* (img_height+margin) + label_padding

    # Create a new blank image with a white background
    grid_img = Image.new('RGB', (grid_width, grid_height), 'white')
    draw = ImageDraw.Draw(grid_img)

    # Load a default fon
    try:
        font= ImageFont.truetype("times_new_roman.ttf", label_font_size)
    except:
        font = ImageFont.load_default()
    # Place each image in the correct position based on the labels
    for i in range(len(images)):
        for j in range(len(images[i])):
            
            
            img = images[i][j]
            print("\t",img)
            # Get the position from labels
            row_idx = i
            col_idx = j
            # Calculate the position where the image will be pasted
            x = col_idx * (img_width +margin) + label_padding +horiz_margin
            y = row_idx * (img_height+margin) + label_padding
            # Resize the image if it's not the correct size
            img = img.resize(image_size)
            # Paste the image into the grid
            grid_img.paste(img, (x, y))
            print(x,y)

    # Draw labels for columns
    for col_idx,label in enumerate(col_labels):
        x = col_idx * (img_width+margin) + label_padding + img_width // 2 +horiz_margin //2
        y = margin  # Position for the column label
        draw.text((x, y), str(label), fill='black', font=font, anchor='mm')

    # Draw labels for rows
    for row_idx,label in enumerate(row_labels):
        x = horiz_margin  # Position for the row label
        y = row_idx * (img_height+margin) + label_padding + img_height // 2
        draw.text((x, y), str(label), fill='black', font=font, anchor='mm')

    return grid_img


from collections import deque
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

class PerPromptStatTracker:
    r"""
    Class for tracking statistics per prompt. Mainly used to calculate advantage for the DPPO algorithm

    Args:
        buffer_size (`int`):
            Size of the buffer to keep for each prompt.
        min_count (`int`):
            Minimum number of samples to keep in the buffer before calculating the mean and std.
    """

    def __init__(self, buffer_size, min_count):
        self.buffer_size = buffer_size
        self.min_count = min_count
        self.stats = {}

    def update(self, prompts, rewards):
        prompts = np.array(prompts)
        rewards = np.array(rewards)
        unique = np.unique(prompts)
        advantages = np.empty_like(rewards)
        for prompt in unique:
            prompt_rewards = rewards[prompts == prompt]
            if prompt not in self.stats:
                self.stats[prompt] = deque(maxlen=self.buffer_size)
            self.stats[prompt].extend(prompt_rewards)

            if len(self.stats[prompt]) < self.min_count:
                mean = np.mean(rewards)
                std = np.std(rewards) + 1e-6
            else:
                mean = np.mean(self.stats[prompt])
                std = np.std(self.stats[prompt]) + 1e-6
            advantages[prompts == prompt] = (prompt_rewards - mean) / std

        return advantages

    def get_stats(self):
        return {k: {"mean": np.mean(v), "std": np.std(v), "count": len(v)} for k, v in self.stats.items()}
